#! python3  # noqa: E265

"""
    Model definition for entwine parameters and related sugar.
"""

# ##################################
# ########## Libraries #############
# ##################################

# Standard library
import json
from pathlib import Path
from sys import version_info

# project
from gpf_common_point_cloud.enums.entwine_enums import EntwineDataType

# Imports depending on Python version
if version_info[1] < 11:
    from typing_extensions import Self
else:
    from typing import Self


# ##################################
# ########## Classes ###############
# ##################################


class GpfEntwineParams:
    """Object definition for entwine parameters."""

    input: str | None = None
    output: str | None = None
    srs: str | None = None
    reprojection: dict | None = None
    dataType: EntwineDataType | None = None
    span: str | None = None
    bounds: list | None = None
    noOriginId: bool | None = False
    absolute: bool | None = None
    scale: int | list | None = None
    maxNodeSize: int | None = None
    minNodeSize: int | None = None
    laz_14: bool | None = False
    limit: int | None = None
    threads: int | None = None
    subset: dict | None = None
    deep: bool | None = None
    tmp: str | None = None

    # internal
    json_ref_path: Path
    loaded_from_json: bool

    @classmethod
    def from_dict(cls, val: dict) -> Self:
        """Load object from a dict.

        Args
            val (dict): dict value to load

        Return
            Self: object with attributes filled from dict.

        Example:

            .. code-block:: python

                entwine_params = GpfEntWineParams.from_dict(
                                        " {"entwine_params": {
                          "dataType": "laszip",
                          "span": 128,
                          "bounds": [
                            0,
                            0,
                            0,
                            100,
                            100,
                            100
                          ],
                          "noOriginId": false,
                          "absolute": true,
                          "scale": [
                            0.1,
                            0.1,
                            0.025
                          ],
                          "maxNodeSize": 30000000,
                          "minNodeSize": 20000000,
                          "laz_14": false
                        }
                    )
        """

        res = cls()

        if "dataType" in val and val["dataType"]:
            res.dataType = val["dataType"]
        if "span" in val:
            res.span = val["span"]
        if "bounds" in val:
            res.bounds = val["bounds"]
        if "absolute" in val:
            res.absolute = val["absolute"]
        if "noOriginId" in val:
            res.noOriginId = val["noOriginId"]
        if "scale" in val:
            res.scale = val["scale"]
        if "maxNodeSize" in val:
            res.maxNodeSize = val["maxNodeSize"]
        if "minNodeSize" in val:
            res.minNodeSize = val["minNodeSize"]
        if "laz_14" in val:
            res.laz_14 = val["laz_14"]
        if "limit" in val:
            res.limit = val["limit"]
        if "threads" in val:
            res.threads = val["threads"]
        if "input" in val:
            res.input = val["input"]
        if "output" in val:
            res.output = val["output"]
        if "deep" in val:
            res.deep = val["deep"]
        if "tmp" in val:
            res.tmp = val["tmp"]

        return res

    def to_json(self, **kwargs) -> str:
        """Supersedes json.dumps using the dictionary returned by to_dict().

        Returns:
            str: object serialized as JSON string

        Example:

            .. code-block:: python

                from pathlib import Path

                # create config file
                out_file = Path("/tmp/config.json")
                out_file.parent.mkdir(parents=True, exist_ok=True)

                # write into the file passing extra parameters to json.dumps
                with out_file.open("w", encoding="UTF8") as wf:
                    wf.write(entwine_param.to_json(sort_keys=True))

        """

        obj_as_dict = self.to_dict()

        # handle cases when the full exception is passed
        if isinstance(obj_as_dict.get("trace"), Exception):
            exc = obj_as_dict.get("trace")
            if isinstance(exc.args, tuple) and len(exc.args):
                exc = exc.args[0]
            else:
                exc = getattr(exc, "message", repr(exc))

            obj_as_dict["trace"] = exc

        return json.dumps(obj_as_dict, **kwargs)

    def to_dict(self) -> dict:
        """Convert dataclass object into dictionary. Supersedes dataclass.asdict to \
        handle enum value.

        Returns:
            dict: object as dictionary

        Example:

            .. code-block:: python

                entwine_param = GpfEntwineParam(
                    "dataType": "laszip",
                    ...
                )
                print(entwine_param.to_dict())


        """

        res = {}

        if self.limit:
            res["limit"] = self.limit
        if self.threads:
            res["threads"] = self.threads
        if self.dataType:
            res["dataType"] = self.dataType
        if self.input:
            res["input"] = self.input
        if self.output:
            res["output"] = self.output
        if self.span:
            res["span"] = self.span
        if self.bounds:
            res["bounds"] = self.bounds
        if self.noOriginId:
            res["noOriginId"] = self.noOriginId
        if self.scale:
            res["scale"] = self.scale
        if self.maxNodeSize:
            res["maxNodeSize"] = self.maxNodeSize
        if self.minNodeSize:
            res["minNodeSize"] = self.minNodeSize
        if self.laz_14:
            res["laz_14"] = self.laz_14
        if self.srs:
            res["srs"] = self.srs
        if self.reprojection:
            res["reprojection"] = self.reprojection
        if self.subset:
            res["subset"] = self.subset
        if self.deep:
            res["deep"] = self.deep
        if self.absolute:
            res["absolute"] = self.absolute
        if self.tmp:
            res["tmp"] = self.tmp
        return res


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass
