#! python3  # noqa: E265

"""
    Helpers to use entwine
"""

# Standard library
import logging
import math
import subprocess
from collections.abc import Iterable
from pathlib import Path

# 3rd party
from gpf_entrepot_toolbelt.orchestrator.status import Status
from gpf_entrepot_toolbelt.utils.check_path import check_file_extension_in_list

# project
from gpf_common_point_cloud.enums.entwine_enums import EntwineDataType, EntwineStep
from gpf_common_point_cloud.utils.laslaz_utils import (
    MAXX_FIELD,
    MAXY_FIELD,
    MINX_FIELD,
    MINY_FIELD,
    READER_NAME,
)
from gpf_common_point_cloud.utils.pdal_utils import pdal_info_from_las

# logs
logger = logging.getLogger(__name__)

DEFAULT_LIMIT = 4
DEFAULT_THREAD = 4
DEFAULT_SPAN = 256
DEFAULT_DATATYPE = EntwineDataType.LASZIP
COUNT_FIELD = "count"
TILES_FIELD = "tiles"
BBOX_FIELD = "bbox"
MAX_TILE_PER_SUBSET: int = 4
MAX_COUNT_PER_SUBSET: int = 500000
MAX_SUBSETS_LEVEL: int = 8


# #############################################################################
# ########## Functions #############
# ##################################


def run_entwine(config_file: Path, step: EntwineStep) -> Status:
    """Execute entwine

    Args:
        config_file (Path): Path to the config file must be used
        step (EntwineStep): step we want to run (build or merge)
    Returns:
        Status:
            SUCCESS if execution OK
            TECHNICAL_ERROR otherwise
    """

    try:
        command = f"entwine {step} -c {config_file}"
        logger.info(f"Lancement de la commande {command}")
        run_result = subprocess.run(command, shell=True, capture_output=True, text=True)

        logger.info(f" Code retour : {run_result.returncode}")
        if run_result.returncode != 0:
            logger.user_error("Erreur lors de l'execution d'entwine")
            logger.error(f" Stdout : {run_result.stdout}")
            logger.error(f" Stderr : {run_result.stderr}")

            return Status.FAILURE

    except Exception as error:
        logger.user_error("Erreur lors de l'execution d'entwine")
        logger.error(error, stack_info=True)
        return Status.TECHNICAL_ERROR

    return Status.SUCCESS


def determine_subsets_number(
    laslaz_dir: Path,
    ignored_file_extensions: Iterable = ["md5"],
) -> int:
    """Function that determine how many subset are needed to avoid

    Args:
        laslaz_dir (Path): path to the upload dir containing las/laz file to compute
        ignored_file_extensions (Iterable, optional): List of ignored file extensions. \
            Defaults to ("md5").

    Returns:
        int: subset number, or -1 if an error occured
    """
    tiles: list[dict] = []
    global_bbox: dict = {
        MINX_FIELD: None,
        MINY_FIELD: None,
        MAXX_FIELD: None,
        MAXY_FIELD: None,
    }
    global_count: int = 0

    # Browse files from upload
    for file in laslaz_dir.rglob("*"):
        if file.is_dir():
            continue

        # Ignored file with extension in list ignored_file_extensions
        if check_file_extension_in_list(
            filepath=file, extensions_list=ignored_file_extensions
        ):
            logger.info(f"Fichier ignoré : {file}.")
            continue

        logger.user_info(f"Récupération des informations de {file}")

        # Get informations from PDAL
        metadata: dict = pdal_info_from_las(file)
        if metadata is None:
            return -1
        logger.debug(f"Métadonnée PDAL : {metadata}.")
        metadata_reader: dict = metadata[READER_NAME]

        xmin: float = float(metadata_reader[MINX_FIELD])
        ymin: float = float(metadata_reader[MINY_FIELD])
        xmax: float = float(metadata_reader[MAXX_FIELD])
        ymax: float = float(metadata_reader[MAXY_FIELD])
        count = int(metadata_reader[COUNT_FIELD])

        # Append information to subset
        tiles.append(
            {
                MINX_FIELD: xmin,
                MINY_FIELD: ymin,
                MAXX_FIELD: xmax,
                MAXY_FIELD: ymax,
                COUNT_FIELD: count,
            }
        )

        if global_bbox[MINX_FIELD] is None or global_bbox[MINX_FIELD] > xmin:
            global_bbox[MINX_FIELD] = xmin

        if global_bbox[MINY_FIELD] is None or global_bbox[MINY_FIELD] > ymin:
            global_bbox[MINY_FIELD] = ymin

        if global_bbox[MAXX_FIELD] is None or global_bbox[MAXX_FIELD] < xmax:
            global_bbox[MAXX_FIELD] = xmax

        if global_bbox[MAXY_FIELD] is None or global_bbox[MAXY_FIELD] < ymax:
            global_bbox[MAXY_FIELD] = ymax

        global_count += count

    # Construct initial subset
    inital_subset: dict = {
        BBOX_FIELD: global_bbox,
        COUNT_FIELD: global_count,
        TILES_FIELD: tiles,
    }

    # Get the maximum necessary subsets level
    subset_level = _subsets_level_recursive(inital_subset, 0)
    logger.info(f"Niveau maximum de subset : {subset_level}")

    # Convert to subsets number
    return int(math.pow(4, subset_level))


def _subsets_level_recursive(subset: dict, subsets_level: int) -> int:
    """Recursive function that :
         determine if a subset is small enough to deal with pod capacity
         if not, split the subset into 4 and recurse

    Args:
        subset (dict): the subset to check
        subsets_level (int): the current subset level

    Returns:
        int: the final subset level for this part
    """

    logger.info(f"Traitement du subset {subset[BBOX_FIELD]}, de niveau {subsets_level}")
    logger.debug(f" Nombre de tuiles : {len(subset[TILES_FIELD])}")
    logger.debug(f" Nombre de points : {subset[COUNT_FIELD]}")

    if (
        len(subset[TILES_FIELD]) <= MAX_TILE_PER_SUBSET
        or subset[COUNT_FIELD] <= MAX_COUNT_PER_SUBSET
        or subsets_level >= MAX_SUBSETS_LEVEL
    ) and subsets_level != 0:
        return subsets_level

    minx: float = subset[BBOX_FIELD][MINX_FIELD]
    maxx: float = subset[BBOX_FIELD][MAXX_FIELD]
    miny: float = subset[BBOX_FIELD][MINY_FIELD]
    maxy: float = subset[BBOX_FIELD][MAXY_FIELD]

    lengthx: float = maxx - minx
    lengthy: float = maxy - miny

    gap: float = abs(lengthx - lengthy) / 2.0
    if lengthx > lengthy:
        miny -= gap
        maxy += gap
    else:
        minx -= gap
        maxx += gap

    meanx: float = (minx + maxx) / 2.0
    meany: float = (miny + maxy) / 2.0

    subset1: dict = _init_empty_subset(minx, miny, meanx, meany)
    subset2: dict = _init_empty_subset(meanx, miny, maxx, meany)
    subset3: dict = _init_empty_subset(minx, meany, meanx, maxy)
    subset4: dict = _init_empty_subset(meanx, meany, maxx, maxy)

    for tile in subset[TILES_FIELD]:
        if tile[MINX_FIELD] < meanx:
            if tile[MINY_FIELD] < meany:
                subset1[TILES_FIELD].append(tile)
                subset1[COUNT_FIELD] += tile[COUNT_FIELD]

            if tile[MAXY_FIELD] > meany:
                subset2[TILES_FIELD].append(tile)
                subset2[COUNT_FIELD] += tile[COUNT_FIELD]

        if tile[MAXX_FIELD] > meanx:
            if tile[MINY_FIELD] < meany:
                subset3[TILES_FIELD].append(tile)
                subset3[COUNT_FIELD] += tile[COUNT_FIELD]

            if tile[MAXY_FIELD] > meany:
                subset4[TILES_FIELD].append(tile)
                subset4[COUNT_FIELD] += tile[COUNT_FIELD]

    subsets_level += 1

    return max(
        [
            _subsets_level_recursive(subset1, subsets_level),
            _subsets_level_recursive(subset2, subsets_level),
            _subsets_level_recursive(subset3, subsets_level),
            _subsets_level_recursive(subset4, subsets_level),
        ]
    )


def _init_empty_subset(minx: float, miny: float, maxx: float, maxy: float):
    """Init a new empty subset with bounds

    Args:
        minx (float): Minimum X for subset bounds
        miny (float): Minimum Y for subset bounds
        maxx (float): Maximum X for subset bounds
        maxy (float): Maximum Y for subset bounds

    Returns:
        dict: the new subset
    """

    return {
        BBOX_FIELD: {
            MINX_FIELD: minx,
            MINY_FIELD: miny,
            MAXX_FIELD: maxx,
            MAXY_FIELD: maxy,
        },
        TILES_FIELD: [],
        COUNT_FIELD: 0,
    }
