#! python3  # noqa: E265

"""Helper to use pdal commands"""

# -- IMPORTS

# standard lib
import json
import logging
import os
import re
import secrets
from pathlib import Path

# 3rd party
from gpf_entrepot_toolbelt.orchestrator.models import GpfExtent
from gpf_entrepot_toolbelt.orchestrator.status import Status
from gpf_entrepot_toolbelt.utils.geometry_utils import transform_coords_to_4326

# package
from gpf_common_point_cloud.enums.entwine_enums import EntwineStep
from gpf_common_point_cloud.models import GpfEntwineParams
from gpf_common_point_cloud.utils.entwine_utils import run_entwine

# logs
logger = logging.getLogger(__name__)

EPT_HEAD_FILE_NAME = "ept.json"
EPT_HEAD_FILE_SUBSET_REGEX = r"^ept-\d+.json$"

EPT_BUILD_FILE_NAME = "ept-build.json"
EPT_BUILD_FILE_SUBSET_REGEX = r"^ept-build-\d+.json$"

EPT_DATA_DIR = "ept-data"
EPT_DATA_FILE_SUBSET_REGEX = r"^ept-data/\d+-\d+-\d+-\d+-\d+.laz$"

EPT_HIERARCHY_DIR = "ept-hierarchy"
EPT_HIERARCHY_FILE_NAME = "0-0-0-0.json"
EPT_HIERARCHY_FILE_SUBSET_REGEX = r"^ept-hierarchy/0-0-0-0-\d+.json$"

EPT_SOURCE_FILE_SUBSET_REGEX = r"^ept-sources/manifest-\d+.json$"


def check_ept_head(
    work_dir: Path, upload_dir: Path, expected_srs: str
) -> tuple[Status, Path | None]:
    """
    Find the EPT pyramid head file en check it

    Args:
        work_dir (Path): Path of a workdir to write temporary file
        upload_dir (Path): Path of the upload data
        expected_srs (str): Expected EPT SRS

    Returns:
        Status: status of the check
        Path: Path to the EPT head file
    """

    logger.user_info(f"Recherche du fichier {EPT_HEAD_FILE_NAME} de la pyramide")

    # Find EPT head
    ept_head_list: list = list(upload_dir.rglob(EPT_HEAD_FILE_NAME))

    if len(ept_head_list) == 0:
        logger.user_error(f"Aucun fichier {EPT_HEAD_FILE_NAME} n'a été trouvé.")
        return Status.FAILURE, None

    if len(ept_head_list) > 1:
        logger.user_error(f"Plusieurs fichiers {EPT_HEAD_FILE_NAME} ont été trouvés.")
        return Status.FAILURE, None

    ept_head_file: Path = ept_head_list.pop(0)

    check_entwine_info_result: Status = check_entwine_info(
        work_dir, ept_head_file, expected_srs=expected_srs
    )
    if check_entwine_info_result != Status.SUCCESS:
        return check_entwine_info_result, None

    return Status.SUCCESS, ept_head_file


def check_ept_hierarchy(ept_dir: Path) -> Status:
    """
    Check the hierarchy of the pyramid

    Args:
        ept_dir (Path): Path to the EPT directory

    Returns:
        Status: status of the check
    """

    logger.user_info(f"Vérification du fichier {EPT_HIERARCHY_FILE_NAME}")

    result: Status = Status.SUCCESS

    # Get hierarchy file
    ept_hierarchy_file = ept_dir / EPT_HIERARCHY_DIR / EPT_HIERARCHY_FILE_NAME
    logger.info(f"Lecture du fichier de hierarchy {ept_hierarchy_file}")
    ept_hierarchy: dict = json.load(open(ept_hierarchy_file.resolve()))

    # Get data file list
    ept_source_dir = ept_dir / EPT_DATA_DIR
    logger.info(f"Lecture des données dans {ept_source_dir}")
    ept_data_list: list[Path] = list(ept_source_dir.resolve().rglob("*"))

    for ept_data_file in ept_data_list:
        logger.debug(f"Vérification du fichier {ept_data_file}")
        ept_data: str = ept_data_file.stem

        try:
            ept_hierarchy.pop(ept_data)

        except KeyError:
            logger.user_error(
                f"Le fichier '{ept_data}' du fichier de hiérarchie n'est pas présent dans la livraison"
            )
            result = Status.FAILURE

    if len(ept_hierarchy) != 0:
        ept_hierarchy_as_str: str = ",".join(ept_hierarchy)
        logger.user_error(
            f"Le(s) fichier(s) : '{ept_hierarchy_as_str}' ne sont pas présent dans le fichier de hiérarchie"
        )
        result = Status.FAILURE

    return result


def check_entwine_info(
    work_dir: Path, ept_head_file: Path, expected_srs: str
) -> Status:
    """
    Execute entwine info command and :
     - check the result
     - the SRS consistency
     - log the EPT version

    Args:
        work_dir (Path): Path of a workdir to write temporary file
        ept_head_file (Path): Path to the EPT head file
        expected_srs (str): Expected EPT SRS

    Returns:
        Status: status of the check
    """

    logger.user_info(f"Vérification des infos de {ept_head_file.name}")

    info_params: GpfEntwineParams = GpfEntwineParams()
    info_params.input = str(ept_head_file.resolve())
    info_params.deep = True
    logger.info(f"Paramètres de entwine : {info_params.to_dict()}")

    entwine_config_file: Path = work_dir / str(secrets.randbits(128))
    with entwine_config_file.open("w", encoding="UTF8") as ecf:
        ecf.write(info_params.to_json())
    run_entwine_result: Status = run_entwine(entwine_config_file, EntwineStep.INFO)
    os.remove(entwine_config_file)
    if run_entwine_result != Status.SUCCESS:
        return run_entwine_result

    logger.info(f"Ouverture du fichier {ept_head_file}")
    ept_head: dict = json.load(open(ept_head_file.resolve()))

    ept_version: str = ept_head["version"]
    logger.user_info(f"Version de l'EPT : {ept_version}")

    srs_found = f"{ept_head['srs']['authority']}:{ept_head['srs']['horizontal']}"
    logger.info(f"SRS de la pyramide : {srs_found}")
    if srs_found != expected_srs:
        logger.user_error(
            f"Le SRS de la pyramide EPT ({srs_found}) "
            f"n'est pas celui déclaré dans la livraison ({expected_srs})"
        )
        return Status.FAILURE

    return Status.SUCCESS


def get_extent(ept_head_file: Path) -> GpfExtent | None:
    """
    Read bounds from EPT head file and return it as GPF extent in 4326

    Args:
        ept_head_file (Path): path to ept head file

    Returns:
        GpfExtent: the generated extent
    """

    ept_head_json: dict = json.load(open(ept_head_file.resolve()))
    bounds: list[float] = ept_head_json["bounds"]
    srs_ept = get_srs(ept_head_file)
    logger.info(f"Bounds du fichier EPT (en {srs_ept}) : {bounds}")

    coords_4326: list[float] = transform_coords_to_4326(
        [bounds[0], bounds[1], bounds[3], bounds[4]], srs_ept
    )
    logger.info(f"Bounds du fichier EPT (en EPSG:4326) : {coords_4326}")

    if coords_4326 is None:
        return None

    return GpfExtent(
        tuple[float]([coords_4326[2], coords_4326[0], coords_4326[3], coords_4326[1]])
    )


def get_srs(ept_head_file: Path) -> str | None:
    """
    Read srs from EPT head file and return it as str

    Args:
        ept_head_file (Path): path to ept head file

    Returns:
        str: srs of the EPT
    """

    try:
        ept_head_json: dict = json.load(open(ept_head_file.resolve()))
        return (
            f"{ept_head_json['srs']['authority']}:{ept_head_json['srs']['horizontal']}"
        )
    except (FileNotFoundError, json.JSONDecodeError, KeyError):
        logger.error(f"Impossible de lire le srs dans le fichier ept: {ept_head_file}")
        return None


def clean_pyramid_temporary_files(pyramid_dir: Path):
    """
    Supprime tous les fichier temporaire d'entwine, générés lors de l'utilisation des subsets
    A exécuter après le merge

    Args:
        pyramid_dir (Path): répertoire de la pyramide
    """

    logger.user_info("Nettoyage des fichiers temporaire de la pyramide EPT")

    for file in pyramid_dir.rglob("*"):
        file_path = str(file.relative_to(pyramid_dir))
        logger.debug(f"Vérification du fichier {file_path}")

        if (
            re.match(EPT_HEAD_FILE_SUBSET_REGEX, file_path)
            or re.match(EPT_BUILD_FILE_SUBSET_REGEX, file_path)
            or re.match(EPT_DATA_FILE_SUBSET_REGEX, file_path)
            or re.match(EPT_HIERARCHY_FILE_SUBSET_REGEX, file_path)
            or re.match(EPT_SOURCE_FILE_SUBSET_REGEX, file_path)
        ):
            logger.info(f"Suppression de {file}")
            os.remove(file)
