#! python3  # noqa: E265

"""Helper to use pdal commands"""

# -- IMPORTS

# standard lib
import logging
import subprocess
from pathlib import Path

# 3rd party
import pdal
from gpf_entrepot_toolbelt.orchestrator.status import Status

# project
from gpf_common_point_cloud.models import PdalTileParameters

# logs
logger = logging.getLogger(__name__)

TERRASCAN_ERROR = "Global encoding WKT flag not set for point format 6 - 10"


def pdal_info_from_las(file: Path) -> dict | None:
    """Execute a PDAL pipeline to get informations from a LAS/LAZ file.

    Args:
        file (Path): path of the file

    Returns:
        pdal.Pipeline: the pipeline executed. None, if error occurs
    """
    pipeline = pdal.Reader.las(filename=file.resolve(), count=0).pipeline()
    try:
        pipeline.execute()
    except RuntimeError as re:
        if TERRASCAN_ERROR in str(re):
            logger.user_error(
                "Erreur Terrascan lors de la lecture du fichier LAS : En LAS 1.4 utilisant le Point Data "
                'Record Format 6 - 10, le flag de Global Encoding (GE) doit être renseigné à "WKT", '
                "sinon une erreur est levée.",
            )
        else:
            logger.user_error(
                "Erreur lors de la lecture du fichier via PDAL",
            )
        logger.error(
            re,
            stack_info=True,
        )
        return None

    return pipeline.metadata["metadata"]


def pdal_merge(files_to_merge: list[Path], output_file_path: Path) -> Status:
    """
    Uses pdal merge to combine input files into a single output file
    Args:
        files_to_merge (list[Path]): input files paths
        output_file_path (Path): output file path

    Returns:
        Status: SUCCESS if the job is ok, FAILURE otherwise
    """
    logger.info(f"pdal merge vers le fichier {output_file_path}")
    try:
        if not files_to_merge:
            logger.warning("La liste des fichiers à fusionner est vide, rien à faire")
            return Status.SUCCESS

        files_to_process: str = " ".join([str(tile) for tile in files_to_merge])
        command: str = f"pdal merge --verbose=2 {files_to_process} {output_file_path}"
        logger.info("Lancement de la commande pdal merge")
        logger.debug(command)
        run_result = subprocess.run(command, shell=True, capture_output=True, text=True)

        if run_result.returncode != 0:
            logger.user_error("Erreur lors de l'execution de pdal merge")
            logger.error(f"Code retour : {run_result.returncode}")
            logger.error(f"Stdout : {run_result.stdout}")
            logger.error(f"Stderr : {run_result.stderr}")
            return Status.FAILURE

        return Status.SUCCESS
    except Exception as error:
        logger.user_error("Erreur lors de l'execution de pdal merge")
        logger.error(error, stack_info=True)
        return Status.FAILURE


def pdal_tile_las(
    pdal_tile_params: PdalTileParameters, input_dir: Path, output_dir: Path
) -> Status:
    """
    Produce tiling from inputs laz files
    Args:
        pdal_tile_params (PdalTileParameters): parameters from the configuration file
        input_dir (Path): POINT-CLOUD las/laz input folder
        output_dir (Path): POINT-CLOUD laz output folder

    Returns:
        Status: SUCCESS if the job is ok, FAILURE otherwise
    """
    try:
        in_pattern: str = f'"{input_dir}/*"'
        command: list = [
            "pdal",
            "tile",
            "--writers.las.forward",
            "header",
            "--writers.las.minor_version",
            "4",
            "-i",
            in_pattern,
            "-o",
            f'"{output_dir}/tile_#.laz"',
        ]

        command.extend([f'--out_srs="{pdal_tile_params.out_srs}"'])
        command.extend([f'--writers.las.a_srs="{pdal_tile_params.out_srs}"'])
        command.extend(["--origin_x", str(pdal_tile_params.origin_x)])
        command.extend(["--origin_y", str(pdal_tile_params.origin_y)])
        command.extend(["--length", str(pdal_tile_params.tile_size)])
        command = " ".join(command)

        logger.info("Lancement de la commande pdal tile")
        logger.debug(command)

        run_result = subprocess.run(command, shell=True, capture_output=True, text=True)

        if run_result.returncode != 0:
            logger.user_error("Erreur lors de l'execution de pdal tile")
            logger.error(f"Code retour : {run_result.returncode}")
            logger.error(f"Stdout : {run_result.stdout}")
            logger.error(f"Stderr : {run_result.stderr}")
            return Status.FAILURE

        return Status.SUCCESS
    except Exception as error:
        logger.user_error("Erreur lors de l'execution de pdal tile")
        logger.error(error, stack_info=True)
        return Status.FAILURE


def pdal_convert_las_to_copc(file_path: Path, output_file_path: Path) -> Status:
    """
    Conversion of files from las/laz to COPC
    Args:
        file_path (Path): las/laz file path to convert
        output_file_path (Path): output file path copc.laz

    Returns:
        Status: SUCCESS if the job is ok, FAILURE otherwise
    """
    logger.info(f"Conversion en COPC de {file_path}")
    try:
        pipeline = pdal.Reader.las(filename=str(file_path)).pipeline()
        pipeline |= pdal.Writer.copc(filename=str(output_file_path))
        pipeline.execute()
        return Status.SUCCESS
    except Exception as error:
        logger.user_error(
            f"Une erreur rencontré lors de la conversion en COPC de {file_path.name}"
        )
        logger.error(error, stack_info=True)
        return Status.FAILURE
