#! python3  # noqa E265
# standard library
import json
import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from gpf_entrepot_toolbelt.orchestrator.status import Status

# 3rd party
from gpf_entrepot_toolbelt.utils.gpf_logger import gpf_logger_script

from gpf_common_point_cloud.utils.vpc_constants import VPC_FILE_NAME

# project
from gpf_common_point_cloud.utils.vpc_utils import (
    build_vpc,
    check_vpc,
    compute_geojson_extent_from_vpc,
)

logger = gpf_logger_script(verbosity=0, title="TestVpcUtils")


# ################################
# ########## Classes #############
# ################################


class TestVpcUtils(unittest.TestCase):
    """Test for core functions"""

    UPLOAD_DIR = "tests/fixtures/vpc"
    VPC_FILE_PATH = UPLOAD_DIR + "/valid.vpc"
    COPC_FILE_1: str = str(
        Path(UPLOAD_DIR).resolve() / "folder/test_0486_6732.copc.laz"
    )
    COPC_FILE_2: str = str(
        Path(UPLOAD_DIR).resolve() / "folder/test_0486_6733.copc.laz"
    )
    COPC_FILE_3: str = str(Path(UPLOAD_DIR).resolve() / "test_0487_6732.copc.laz")
    COPC_FILE_4: str = str(Path(UPLOAD_DIR).resolve() / "test_0487_6733.copc.laz")

    def setUp(self):
        self.maxDiff = None

    def test_check_vpc_ok(self):
        # Given
        vpc_file: Path = Path(self.VPC_FILE_PATH)
        copc_list = [
            self.COPC_FILE_1,
            self.COPC_FILE_2,
            self.COPC_FILE_3,
            self.COPC_FILE_4,
        ]

        # When
        result: Status = check_vpc(vpc_file, copc_list)

        # Then
        self.assertEqual(Status.SUCCESS, result)

    def test_check_vpc_nok_reading_error(self):
        # Given
        vpc_file: Path = Path("tests/fixtures/vpc/reading_error.vpc")
        copc_list = [
            self.COPC_FILE_1,
        ]

        # When
        result: Status = check_vpc(vpc_file, copc_list)

        # Then
        self.assertEqual(Status.FAILURE, result)

    def test_check_vpc_nok_no_features(self):
        # Given
        vpc_file: Path = Path("tests/fixtures/vpc/no_features.vpc")
        copc_list = [
            self.COPC_FILE_1,
        ]

        # When
        result: Status = check_vpc(vpc_file, copc_list)

        # Then
        self.assertEqual(Status.FAILURE, result)

    def test_check_vpc_nok_bad_features(self):
        # Given
        vpc_file: Path = Path("tests/fixtures/vpc/bad_features.vpc")
        copc_list = [
            self.COPC_FILE_1,
        ]

        # When
        result: Status = check_vpc(vpc_file, copc_list)

        # Then
        self.assertEqual(Status.FAILURE, result)

    def test_check_vpc_nok_missing_copc_in_files(self):
        # Given
        vpc_file: Path = Path(self.VPC_FILE_PATH)
        copc_list = [self.COPC_FILE_1, self.COPC_FILE_2, self.COPC_FILE_3]

        # When
        result: Status = check_vpc(vpc_file, copc_list)

        # Then
        self.assertEqual(Status.FAILURE, result)

    def test_check_vpc_nok_missing_copc_in_vpc(self):
        # Given
        vpc_file: Path = Path(self.VPC_FILE_PATH)
        copc_list = [
            self.COPC_FILE_1,
            self.COPC_FILE_2,
            self.COPC_FILE_3,
            self.COPC_FILE_4,
            "LHD_FXX_0488_6732_PTS_C_LAMB93_IGN69.copc.laz",
        ]

        # When
        result: Status = check_vpc(vpc_file, copc_list)

        # Then
        self.assertEqual(Status.FAILURE, result)

    def test_check_vpc_nok_relative(self):
        # Given
        vpc_file: Path = Path("tests/fixtures/vpc/relative.vpc")
        copc_list = [
            "../tile_482092_4390629.copc.laz",
        ]

        # When
        result: Status = check_vpc(vpc_file, copc_list)

        # Then
        self.assertEqual(Status.FAILURE, result)

    def test_compute_global_extent_from_vpc_ok(self):
        # Given
        vpc_path = Path("./tests/fixtures/vpc/index.vpc")

        # When
        result = compute_geojson_extent_from_vpc(vpc_path)
        expected_extent = json.load(
            open(Path("./tests/fixtures/vpc/expected_extent.json"))
        )

        # Then
        self.assertDictEqual(expected_extent, json.loads(json.dumps(result)))

    def test_compute_global_extent_from_vpc_nok_reading_error(self):
        # Given
        vpc_path = Path("./tests/fixtures/vpc/reading_error.vpc")

        # When
        result = compute_geojson_extent_from_vpc(vpc_path)

        # Then
        self.assertIsNone(result)

    def test_compute_global_extent_from_vpc_nok_no_features(self):
        # Given
        vpc_path = Path("./tests/fixtures/vpc/no_features.vpc")

        # When
        result = compute_geojson_extent_from_vpc(vpc_path)

        # Then
        self.assertIsNone(result)

    def test_compute_global_extent_from_vpc_nok_bad_features(self):
        # Given
        vpc_path = Path("./tests/fixtures/vpc/bad_features.vpc")

        # When
        result = compute_geojson_extent_from_vpc(vpc_path)

        # Then
        self.assertIsNone(result)

    @patch("pathlib.Path.glob")
    @patch("subprocess.run")
    def test_build_vpc_ok(self, mock_run, mock_glob):
        mock_glob.return_value = [
            Path("/fake/path/file1.las"),
            Path("/fake/path/file2.laz"),
        ]

        mock_proc = MagicMock(returncode=0)
        mock_proc.communicate.return_value = (b"standard output", b"")
        mock_run.return_value = mock_proc

        work_dir = Path("/fake/work_dir")
        output_dir = Path("/fake/output_dir")

        result = build_vpc(work_dir, output_dir)

        mock_glob.assert_called_once_with("*")
        command_to_run = (
            f"pdal_wrench build_vpc "
            f"--output=/fake/output_dir/{VPC_FILE_NAME} "
            f"/fake/path/file1.las /fake/path/file2.laz"
        )
        mock_run.assert_called_once_with(
            command_to_run, shell=True, capture_output=True, text=True
        )

        self.assertEqual(Status.SUCCESS, result)

    @patch("pathlib.Path.glob")
    @patch("subprocess.run")
    def test_build_vpc_nok_code_return(self, mock_run, mock_glob):
        mock_glob.return_value = [
            Path("/fake/path/file1.las"),
            Path("/fake/path/file2.laz"),
        ]

        mock_proc = MagicMock()
        mock_proc.communicate.return_value = (b"", b"standard error")
        mock_run.return_value = mock_proc

        work_dir = Path("/fake/work_dir")
        output_dir = Path("/fake/output_dir")

        result = build_vpc(work_dir, output_dir)

        mock_glob.assert_called_once_with("*")
        command_to_run = (
            f"pdal_wrench build_vpc "
            f"--output=/fake/output_dir/{VPC_FILE_NAME} "
            f"/fake/path/file1.las /fake/path/file2.laz"
        )
        mock_run.assert_called_once_with(
            command_to_run, shell=True, capture_output=True, text=True
        )

        self.assertEqual(Status.TECHNICAL_ERROR, result)

    @patch("pathlib.Path.glob")
    @patch("subprocess.run")
    def test_build_vpc_nok_exception(self, mock_run, mock_glob):
        mock_glob.return_value = [
            Path("/fake/path/file1.las"),
            Path("/fake/path/file2.laz"),
        ]

        mock_run.side_effect = Exception("Test exception")

        work_dir = Path("/fake/work_dir")
        output_dir = Path("/fake/output_dir")

        result = build_vpc(work_dir, output_dir)

        mock_glob.assert_called_once_with("*")
        command_to_run = (
            f"pdal_wrench build_vpc "
            f"--output=/fake/output_dir/{VPC_FILE_NAME} "
            f"/fake/path/file1.las /fake/path/file2.laz"
        )
        mock_run.assert_called_once_with(
            command_to_run, shell=True, capture_output=True, text=True
        )

        self.assertEqual(Status.TECHNICAL_ERROR, result)

    # ################################
    # ####### Stand-alone run ########
    # ################################
    if __name__ == "__main__":
        unittest.main()
